> module DeBruijn where
> import Prelude hiding (abs)
> import Rank2

Rank 2 type constructor for a de Bruijn representation of lambda-terms.

> data LamF f a = Var a | App (f a) (f a) | Abs (f (Maybe a))
>
> instance Functor2 LamF where
>     ffmap s f (Var a) = Var (f a)
>     ffmap s f (App t1 t2) = App (s f t1) (s f t2)
>     ffmap s f (Abs r) = Abs (s (fmap f) r)

Type of de Bruijn terms over a variable set A.

> type Lam a = Mu2 LamF a

De Bruijn term constructors.

> var :: a -> Lam a
> var a = In (Var a)
>
> app :: Lam a -> Lam a -> Lam a
> app t1 t2 = In (App t1 t2)
>
> abs :: Lam (Maybe a) -> Lam a
> abs r = In (Abs r)

Weakening.

> weak :: Lam a -> Lam (Maybe a)
> weak t = fmap Just t

Parallel substitution.

Step term. 

> newtype RKanLam b = RKanLam (forall c. ((b -> Lam c) -> Lam c))
>
> stepSub :: LamF RKanLam a -> RKanLam a
>
> stepSub (Var a)                         = RKanLam (\sigma -> sigma a)
>
> stepSub (App (RKanLam t1) (RKanLam t2)) = RKanLam (\sigma -> 
>   app (t1 sigma) (t2 sigma))
>
> stepSub (Abs (RKanLam r))               = RKanLam (\sigma -> 
>   abs (r (maybe (var Nothing) (weak . sigma))))

Substitution in general form.

> subst' :: (a -> b) -> Lam a -> RKanLam b
> subst' = it stepSub

Substitution (monad operation).

> subst :: Lam a -> (a -> Lam b) -> Lam b
> subst t sigma = k sigma 
>     where (RKanLam k) = subst' id t

Join operation.

> join' :: (a -> b) -> Lam (Lam a) -> Lam b
> join' f t = k id where (RKanLam k) = subst' (fmap f) t

> join  :: Lam (Lam a) -> Lam a
> join t = subst t id

\comm{ begin comment

Example.

\[
\begin{array}{lll}
    t & = & x (\lambda y. y x) \\
    \sigma & = & [z (\lambda w. z w) / x] \\
\end{array}
\]

> t1 :: Lam Int
> t1 = app (var 0) (abs (app (var Nothing) (var (Just 0))))
>
> sigma :: Int -> Lam String
> sigma 0 = app (var "z") (abs (app (var (Just "z")) (var Nothing)))
>
> t1' :: Lam String
> t1' = subst t1 sigma
 
Printing.

> newtype RKanString b = RKanString ((b -> String) -> String)

> stepPrint :: LamF RKanString a -> RKanString a
>
> stepPrint (Var a)                               = RKanString (\p -> 
>   p a)
>
> stepPrint (App (RKanString t1) (RKanString t2)) = RKanString (\p ->
>   "(" ++ t1 p ++ " " ++ t2 p ++ ")")
>
> stepPrint (Abs (RKanString r))                  = RKanString (\p -> 
>   "\\" ++ r (lift p)) 
>   where lift p Nothing = "0"
>	  lift p (Just a) = "S" ++ p a

> printLam :: Lam a -> (a -> String) -> String 
> printLam t p = k p where (RKanString k) = it stepPrint id t

> instance Show a => Show (Lam a) where
>     show t = printLam t show

end comment }
